import {
  SyntheticEvent,
  useCallback,
  useContext,
  useEffect,
  useMemo,
  useRef,
} from "react";
import {
  NativeSyntheticEvent,
  ReturnKeyTypeOptions,
  TextInput,
  TextInputChangeEventData,
  TextInputFocusEventData,
} from "react-native";

import { FormFieldValue, FormsContext } from "./FormsContext";

type FormValues = { [name: string]: FormFieldValue };

type UseFormAPI<
  FV extends FormValues = FormValues,
  FN extends keyof FV = keyof FV,
> = {
  registerField<F extends FN = FN>(
    fieldName: F,
    opts?: {
      nextField?: Exclude<keyof FV, F>;
      returnKeyType?: ReturnKeyTypeOptions;
      returnSubmitsForm?: boolean;
    },
  ): {
    onBlur: (e: NativeSyntheticEvent<TextInputFocusEventData>) => void;
    onChange: (e: NativeSyntheticEvent<TextInputChangeEventData>) => void;
    onSubmitEditing?: () => void;
    ref: (fieldRef: TextInput | null) => void;
    returnKeyType?: ReturnKeyTypeOptions;
    value?: FormFieldValue;
  };
  handleSubmit(
    wrappedFn: (ev: null | SyntheticEvent<any, Event>, formValues: FV) => void,
  ): () => void;
};

export function useForm<
  FV extends FormValues = FormValues,
  FN extends keyof FV = keyof FV,
>(
  formId: string,
  opts?: {
    initialValues?: Partial<FV>;
  },
): UseFormAPI<FV, FN> {
  const handleSubmitFunctionRef =
    useRef<
      null | ((ev: null | SyntheticEvent<any, Event>, formValues: FV) => void)
    >(null);
  const fieldsRefsRef = useRef<{
    [fieldName: string]: TextInput;
  }>({});

  const { _getFormValues, _getFormFieldValue, _setFormFieldValue } =
    useContext(FormsContext);

  const registerField = useCallback(
    <F extends FN = FN>(
      fieldName: F,
      fieldOpts?: {
        nextField?: Exclude<keyof FV, F>;
        returnKeyType?: ReturnKeyTypeOptions;
        returnSubmitsForm?: boolean;
      },
    ) => {
      return {
        onChange: (
          ev: NativeSyntheticEvent<TextInputChangeEventData>,
        ): void => {
          ev.persist();
          if (ev?.nativeEvent?.text == null) return undefined;
          _setFormFieldValue(
            formId,
            fieldName as string,
            ev.nativeEvent.text as string,
          );
        },
        onBlur: (_ev: NativeSyntheticEvent<TextInputFocusEventData>): void =>
          void 0,
        onSubmitEditing:
          fieldOpts?.nextField != null
            ? () => {
                const ref = Object.entries(fieldsRefsRef.current).find(
                  ([fName]) => fName === fieldOpts.nextField,
                );
                if (ref?.[1] != null) {
                  const inputRef = ref[1];
                  if (
                    inputRef.isFocused() === false &&
                    inputRef?.props?.editable !== false
                  ) {
                    inputRef.focus();
                  }
                }
              }
            : fieldOpts?.returnSubmitsForm &&
              handleSubmitFunctionRef.current != null
            ? () => {
                const formValues = _getFormValues(formId);
                handleSubmitFunctionRef?.current?.(null, formValues as FV);
              }
            : undefined,
        ref: (instance: TextInput | null) => {
          if (instance) {
            fieldsRefsRef.current = {
              ...(fieldsRefsRef?.current || {}),
              [fieldName]: instance,
            };
          }
        },
        returnKeyType:
          fieldOpts?.nextField != null
            ? "next"
            : fieldOpts?.returnKeyType || undefined,
        value: _getFormFieldValue(formId, fieldName as string) || undefined,
      };
    },
    [_getFormFieldValue, _setFormFieldValue, formId],
  );

  const handleSubmit = useCallback(
    (
      wrappedFn: (
        ev: null | SyntheticEvent<any, Event>,
        formValues: FV,
      ) => void,
    ) => {
      handleSubmitFunctionRef.current = wrappedFn;
      return (ev?: SyntheticEvent<any, Event>): void => {
        const formValues = _getFormValues(formId);
        wrappedFn(
          ev as SyntheticEvent<any, Event>,
          formValues as unknown as FV,
        );
        return undefined;
      };
    },
    [_getFormValues, formId],
  );

  useEffect(() => {
    if (opts?.initialValues != null) {
      Object.entries(opts.initialValues).forEach(([k, v]) => {
        _setFormFieldValue(formId, k, v);
      });
    }
  }, []);

  return useMemo(
    () => ({
      registerField,
      handleSubmit,
    }),
    [registerField, handleSubmit],
  );
}
